from typing import Any, Optional
from urllib.error import HTTPError

from requests import Response

from .assets import ModeAnalyticsAsset

RATE_LIMIT_STATUS_CODE = 429


class UnexpectedApiResponseError(Exception):
    """Custom error handling case of unexpected API Response of Mode Analytics"""

    def __init__(self, resource_name: Optional[str], result: Any):
        error_msg = "Could not extract result from API response."
        error_msg += f"resource_name: {resource_name}"
        error_msg += f"result: {result}"
        super().__init__(error_msg)


class MissingPrerequisiteError(Exception):
    """Custom error handling case of missing asset"""

    def __init__(
        self,
        fetched: ModeAnalyticsAsset,
        missing: ModeAnalyticsAsset,
    ):
        error_msg = f"{missing.name} must be provided to fetch {fetched.name}."
        super().__init__(error_msg)


class RateLimitResponseError(Exception):
    """Custom error handling case of RateLimit of Mode Analytics API"""

    def __init__(self):
        error_msg = "Could not fetch result from API response due to RateLimit"
        super().__init__(error_msg)


def check_errors(response: Response):
    """
    Error check from response.
    If error is 429 (RATE LIMIT) raise custom error RateLimiteResponseError.
    """
    try:
        response.raise_for_status()
    except HTTPError as err:
        if err.code == RATE_LIMIT_STATUS_CODE:
            raise RateLimitResponseError()
        raise err



import os
from typing import Optional, overload

from typing_extensions import Literal


@overload
def from_env(key: str, allow_missing: Literal[False]) -> str:
    ...


@overload
def from_env(key: str, allow_missing: bool) -> Optional[str]:
    ...


@overload
def from_env(
    key: str,
) -> str:
    ...


def from_env(key: str, allow_missing: bool = False) -> Optional[str]:
    """Return the value of the given environment variable"""
    value = os.environ.get(key)

    if value:
        return value

    if allow_missing:
        return None

    raise KeyError(f"Missing {key} in ENV variables")



from argparse import ArgumentParser

from castor_extractor.warehouse import bigquery  # type: ignore


def main():
    parser = ArgumentParser()

    parser.add_argument(
        "-c", "--credentials", help="File path to google credentials"
    )
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument(
        "--skip-existing",
        dest="skip_existing",
        action="store_true",
        help="Skips files already extracted instead of replacing them",
    )

    parser.add_argument(
        "--db-allowed",
        nargs="*",
        help="List of gcp projects that should be extracted",
    )
    parser.add_argument(
        "--db-blocked",
        nargs="*",
        help="List of gcp projects that should not be extracted",
    )
    parser.add_argument(
        "--safe-mode",
        "-s",
        help="bigquery safe mode",
        action="store_true",
    )

    parser.set_defaults(skip_existing=False)

    args = parser.parse_args()

    bigquery.extract_all(
        credentials=args.credentials,
        output_directory=args.output,
        skip_existing=args.skip_existing,
        db_allowed=args.db_allowed,
        db_blocked=args.db_blocked,
        safe_mode=args.safe_mode,
    )



from .collection import group_by
from .constants import OUTPUT_DIR
from .env import from_env
from .files import explode, search_files
from .formatter import to_string_array
from .load import load_file
from .object import deep_serialize, getproperty
from .pager import Pager, PagerLogger, PagerStopStrategy
from .retry import RetryStrategy, retry
from .safe import SafeMode, safe_mode
from .store import AbstractStorage, LocalStorage
from .string import string_to_tuple
from .time import current_date, current_datetime, current_timestamp
from .type import Callback, Getter, JsonType, SerializedAsset
from .uri import uri_encode
from .validation import validate_baseurl
from .write import (
    get_output_filename,
    get_summary_filename,
    get_summary_payload,
    write_errors_logs,
    write_json,
    write_summary,
)



from typing import Optional

DETAILS_KEY = "details"
DBNAME_KEYS = ("dbname", "db", "project_id", "project-id")


def get_dbname_from_details(details: dict) -> Optional[str]:
    """
    Extract dbname from details, trying several keys.
    Returns None if all keys failed.
    """
    for key in DBNAME_KEYS:
        if key in details:
            return details.get(key)
    return None



COLUMN_TYPES = {
    "ARRAY",
    "BINARY",
    "BOOLEAN",
    "CUSTOM",
    "DATE",
    "DECIMAL",
    "ENUM",
    "FLOAT",
    "GEOGRAPHY",
    "INTEGER",
    "INTERVAL",
    "OBJECT",
    "STRING",
    "TEXT",
    "TIME",
    "TIMESTAMP",
}

TABLE_TYPES = {"TABLE", "VIEW", "EXTERNAL"}



from enum import Enum
from typing import Dict, Tuple


class WarehouseAsset(Enum):
    """Assets that can be extracted from warehouses"""

    COLUMN = "column"
    DATABASE = "database"
    GRANT_TO_ROLE = "grant_to_role"
    GRANT_TO_USER = "grant_to_user"
    GROUP = "group"
    QUERY = "query"
    ROLE = "role"
    SCHEMA = "schema"
    TABLE = "table"
    USER = "user"
    VIEW_DDL = "view_ddl"


class WarehouseAssetGroup(Enum):
    """Groups of assets that can be extracted together"""

    CATALOG = "catalog"
    ROLE = "role"
    QUERY = "query"
    VIEW_DDL = "view_ddl"


# tuple of supported assets for each group (depends on the technology)
SupportedAssets = Dict[WarehouseAssetGroup, Tuple[WarehouseAsset, ...]]


# shared by all technologies
CATALOG_ASSETS = (
    WarehouseAsset.DATABASE,
    WarehouseAsset.SCHEMA,
    WarehouseAsset.TABLE,
    WarehouseAsset.COLUMN,
)

# shared by technologies supporting queries
QUERIES_ASSETS = (WarehouseAsset.QUERY,)
VIEWS_ASSETS = (WarehouseAsset.VIEW_DDL,)



from enum import Enum
from typing import Dict, Optional

from ....utils import from_env

AUTH_ERROR_MSG = "Need either user and password or token_name and token"


class CredentialsKey(Enum):
    """Value enum object for the credentials"""

    TABLEAU_USER = "user"
    TABLEAU_PASSWORD = "password"
    TABLEAU_TOKEN_NAME = "token_name"
    TABLEAU_TOKEN = "token"
    TABLEAU_SITE_ID = "site_id"
    TABLEAU_SERVER_URL = "server_url"


CREDENTIALS_ENV: Dict[CredentialsKey, str] = {
    CredentialsKey.TABLEAU_USER: "CASTOR_TABLEAU_USER",
    CredentialsKey.TABLEAU_PASSWORD: "CASTOR_TABLEAU_PASSWORD",
    CredentialsKey.TABLEAU_TOKEN_NAME: "CASTOR_TABLEAU_TOKEN_NAME",
    CredentialsKey.TABLEAU_TOKEN: "CASTOR_TABLEAU_TOKEN",
    CredentialsKey.TABLEAU_SITE_ID: "CASTOR_TABLEAU_SITE_ID",
    CredentialsKey.TABLEAU_SERVER_URL: "CASTOR_TABLEAU_SERVER_URL",
}


def get_value(
    key: CredentialsKey, kwargs: dict, optional: bool = False
) -> Optional[str]:
    """
    Returns the value of the given key:
    - from kwargs in priority
    - from ENV if not provided (raises an error if not found in ENV)
    """
    env_key = CREDENTIALS_ENV[key]

    return kwargs.get(key.value) or from_env(env_key, optional)


class CredentialsApi:
    """ValueObject for the credentials"""

    def __init__(
        self,
        *,
        server_url: str,
        site_id: str,
        user: Optional[str],
        password: Optional[str],
        token_name: Optional[str],
        token: Optional[str],
    ):
        credentials = self._get_credentials(user, password, token_name, token)

        self.user = credentials.get(CredentialsKey.TABLEAU_USER)
        self.site_id = site_id
        self.server_url = server_url
        self.password = credentials.get(CredentialsKey.TABLEAU_PASSWORD)
        self.token_name = credentials.get(CredentialsKey.TABLEAU_TOKEN_NAME)
        self.token = credentials.get(CredentialsKey.TABLEAU_TOKEN)

    @staticmethod
    def _get_credentials(
        user: Optional[str],
        password: Optional[str],
        token_name: Optional[str],
        token: Optional[str],
    ) -> dict:
        """Helpers to retrieve credentials,
        if both are given choose user and password authentication method"""
        assert (user and password) or (token_name and token), AUTH_ERROR_MSG

        if user and password:
            return {
                CredentialsKey.TABLEAU_USER: user,
                CredentialsKey.TABLEAU_PASSWORD: password,
            }

        return {
            CredentialsKey.TABLEAU_TOKEN_NAME: token_name,
            CredentialsKey.TABLEAU_TOKEN: token,
        }

    def to_dict(self, hide: bool = False) -> Dict[str, str]:
        safe = (
            CredentialsKey.TABLEAU_USER,
            CredentialsKey.TABLEAU_SITE_ID,
            CredentialsKey.TABLEAU_SERVER_URL,
            CredentialsKey.TABLEAU_TOKEN_NAME,
        )
        unsafe = (CredentialsKey.TABLEAU_PASSWORD, CredentialsKey.TABLEAU_TOKEN)

        def val(k: CredentialsKey, v: str) -> str:
            return "*" + v[-4:] if hide and k in unsafe else v

        return {a.value: val(a, getattr(self, a.value)) for a in safe + unsafe}



import logging
from typing import Iterable, Tuple, Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    validate_baseurl,
    write_json,
    write_summary,
)
from .api import ApiClient, Credentials, explore_names_associated_to_dashboards
from .assets import LookerAsset

BASE_URL = "CASTOR_LOOKER_BASE_URL"
CLIENT_ID = "CASTOR_LOOKER_CLIENT_ID"
CLIENT_SECRET = "CASTOR_LOOKER_CLIENT_SECRET"
ALL_LOOKS = "CASTOR_LOOKER_ALL_LOOKS"

logger = logging.getLogger(__name__)


def _all_looks_from_env():
    parameter = from_env(ALL_LOOKS, allow_missing=True)
    return str(parameter).lower() == "true"


def iterate_all_data(**kwargs) -> Iterable[Tuple[LookerAsset, list]]:
    """Iterate over the extracted Data From looker"""

    base_url = validate_baseurl(kwargs.get("base_url") or from_env(BASE_URL))
    client_id = kwargs.get("client_id") or from_env(CLIENT_ID)
    client_secret = kwargs.get("client_secret") or from_env(CLIENT_SECRET)
    all_looks = kwargs.get("all_looks") or _all_looks_from_env()

    # timeout can be set with env variable however we don't use from_env because it has a default value
    timeout = kwargs.get("timeout")

    client = ApiClient(
        credentials=Credentials(
            base_url=base_url,
            client_id=client_id,
            client_secret=client_secret,
            timeout=timeout,
        )
    )

    logger.info("Extracting users from Looker API")
    yield LookerAsset.USERS, deep_serialize(client.users())

    logger.info("Extracting folders from Looker API")
    yield LookerAsset.FOLDERS, deep_serialize(client.folders())

    logger.info("Extracting looks from Looker API")
    yield LookerAsset.LOOKS, deep_serialize(client.looks(all_looks))

    logger.info("Extracting dashboards from Looker API")
    dashboards = client.dashboards()
    yield LookerAsset.DASHBOARDS, deep_serialize(dashboards)

    logger.info("Extracting lookml models from Looker API")
    lookmls = client.lookml_models()
    yield LookerAsset.LOOKML_MODELS, deep_serialize(lookmls)

    logger.info("Extracting explores from Looker API")
    explore_names = explore_names_associated_to_dashboards(lookmls, dashboards)
    yield LookerAsset.EXPLORES, deep_serialize(client.explores(explore_names))

    logger.info("Extracting connections from Looker API")
    connections = client.connections()
    yield LookerAsset.CONNECTIONS, deep_serialize(connections)

    logger.info("Extracting projects from Looker API")
    projects = client.projects()
    yield LookerAsset.PROJECTS, deep_serialize(projects)


def extract_all(**kwargs) -> None:
    """Extract Data From looker and store it locally in files under the output_directory"""
    output_directory = kwargs.get("output_directory") or from_env(OUTPUT_DIR)
    base_url = validate_baseurl(kwargs.get("base_url") or from_env(BASE_URL))

    ts = current_timestamp()

    for key, data in iterate_all_data(**kwargs):
        filename = get_output_filename(key.value, output_directory, ts)
        write_json(filename, data)

    write_summary(output_directory, ts, base_url=base_url)



from typing import Callable, List, Optional

from tqdm import tqdm  # type: ignore

from ..assets import QlikAsset
from .constants import APP_EXTERNAL_ID_KEY, SCOPED_ASSETS
from .engine import EngineApiClient
from .rest import RestApiClient

ListedData = List[dict]


class MissingAppsScopeError(Exception):
    """
    Error to be raised when the scope on apps was required and not provided
    """

    def __init__(self, asset: QlikAsset):
        error_msg = f"App ids must be provided to fetch {asset}."
        super().__init__(error_msg)


def _include_app_external_id(
    data: ListedData, app_external_id: str
) -> ListedData:
    _data = data.copy()
    for element in _data:
        element.update({"app_external_id": app_external_id})
    return _data


def _fetch_children_on_apps(
    apps: ListedData, fetch_callback: Callable, display_progress: bool
) -> ListedData:
    all_apps_data: ListedData = list()
    apps_iterator = apps if not display_progress else tqdm(apps)
    for app in apps_iterator:
        app_id = app[APP_EXTERNAL_ID_KEY]
        data = fetch_callback(app_id)
        data = _include_app_external_id(data, app_id)
        all_apps_data.extend(data)
    return all_apps_data


class QlikMasterClient:
    """
    Qlik master client acts as a wrapper class on top of Qlik RestApiClient and
    EngineApiClient to fetch assets regardless of the underlying API.
    """

    def __init__(
        self,
        server_url: str,
        api_key: str,
        except_http_error_statuses: Optional[List[int]] = None,
        display_progress: bool = True,
    ):
        self._server_url = server_url
        self._api_key = api_key
        self.display_progress = display_progress

        self.rest_api_client = RestApiClient(
            server_url=self._server_url,
            api_key=self._api_key,
            except_http_error_statuses=except_http_error_statuses,
        )

        self.engine_api_client = EngineApiClient(
            server_url=self._server_url, api_key=self._api_key
        )

    def _fetch_lineage(self, apps: ListedData) -> ListedData:
        callback = self.rest_api_client.data_lineage
        return _fetch_children_on_apps(apps, callback, self.display_progress)

    def _fetch_measures(self, apps: ListedData) -> ListedData:
        callback = self.engine_api_client.measures
        return _fetch_children_on_apps(apps, callback, self.display_progress)

    def fetch(
        self, asset: QlikAsset, *, apps: Optional[ListedData] = None
    ) -> ListedData:
        """
        Given a QlikAsset, returns the corresponding data using the
        appropriate client.

        Note:
            QlikAsset.LINEAGE and QlikAsset.MEASURES must be provided a
            scope on app_ids
        """
        if asset in SCOPED_ASSETS and not apps:
            raise MissingAppsScopeError(asset)

        if asset == QlikAsset.MEASURES:
            assert apps  # can't be False as we priorly checked
            return self._fetch_measures(apps)

        if asset == QlikAsset.LINEAGE:
            assert apps  # can't be False as we priorly checked
            return self._fetch_lineage(apps)

        return self.rest_api_client.get(asset)
